#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import json
import uuid
import time
import base64

import appier

HELLO_WORLD_B64 = "SGVsbG8gV29ybGQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAQAAAAEAAABAAQAAAA\
AAAAAAAABDYWxpYnJpAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACQAAAAMAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
AAAAAAAAAwAAABIZWxsbyBXb3JsZAA="

VALID_OPTIONS = set(
    [
        "scale",
        "quality",
        "email_address",
        "email_receivers",
        "email_receiver",
        "email_override",
    ]
)


class NodeController(appier.Controller):
    @appier.route("/nodes", "GET", json=True)
    @appier.ensure(token="admin")
    def list(self):
        return self.owner.nodes

    @appier.route("/nodes/<str:id>", "POST", json=True)
    @appier.ensure(token="admin")
    def create(self, id):
        node = appier.get_object()
        self.owner.nodes[id] = node

    @appier.route("/nodes/<str:id>", "GET", json=True)
    @appier.ensure(token="admin")
    def show(self, id):
        return self.owner.nodes[id]

    @appier.route("/nodes/<str:id>/jobs", "GET", json=True)
    @appier.ensure(token="admin")
    def jobs(self, id):
        self.request.set_content_type("application/json")
        for value in appier.header_a():
            yield value
        for value in self.wait_jobs(id):
            yield value

    @appier.route("/nodes/<str:id>/jobs_peek", "GET", json=True)
    @appier.ensure(token="admin")
    def jobs_peek(self, id):
        jobs = self.owner.jobs.get(id, [])
        return jobs

    @appier.route("/nodes/<str:id>/jobs/<str:job_id>/result", "POST", json=True)
    @appier.ensure(token="admin")
    def job_result(self, id, job_id):
        data_path = appier.conf("DATA_PATH", "./data")
        job_path = os.path.join(data_path, job_id)

        payload = appier.get_object()
        data = payload.get("data", dict())
        files = data.pop("files", [])

        job_info = self.owner.jobs_info[job_id]
        if not id == job_info["node_id"]:
            raise appier.OperationalError("Node ID mismatch")

        if (payload or files) and not os.path.exists(job_path):
            os.makedirs(job_path)

        if payload:
            payload_s = json.dumps(payload).encode("utf-8")
            with open(os.path.join(job_path, "payload.json"), "wb") as _file:
                _file.write(payload_s)

        for file in files:
            _file = appier.File(file)
            _file.save(path=os.path.join(job_path, _file.file_name))

        job_info.update(status="finished", finish_time=time.time(), result=payload)

    @appier.route("/nodes/<str:id>/print", ("GET", "POST"), json=True)
    @appier.ensure(token="admin")
    def print_default(self, id):
        data = self.field("data", None)
        data_b64 = self.field("data_b64", None)
        name = self.field("name", None)
        type = self.field("type", None)
        format = self.field("format", None)
        options = self.field("options", None, cast=dict)

        appier.verify(
            data or data_b64,
            message="Either data or data_b64 fields must be provided",
            code=400,
        )
        appier.verify(
            not (data and data_b64),
            message="Only one of data or data_b64 fields must be provided",
            code=400,
        )

        job_id = str(uuid.uuid4())
        name = name or job_id
        if data:
            data_b64 = base64.b64encode(
                appier.legacy.bytes(data, encoding="utf-8")
            ).decode("utf-8")

        job_info = dict(id=job_id, name=name, node_id=id, data_length=len(data_b64))
        if type:
            job_info["type"] = type
        if format:
            job_info["format"] = format
        if options:
            job_info["options"] = dict(
                (k, v) for k, v in options.items() if k in VALID_OPTIONS
            )
        self.owner.jobs_info[job_id] = job_info

        # creates a copy of the job info as starting
        # point for the job structure and then adds
        # the "heavy" data (base64 encoded) to it
        job = dict(job_info)
        job["data_b64"] = data_b64
        jobs = self.owner.jobs.get(id, [])
        jobs.append(job)
        self.owner.jobs[id] = jobs
        appier.notify("jobs:%s" % id)

        job_info.update(status="queued", queued_time=time.time())
        return job_info

    @appier.route("/nodes/<str:id>/print", "OPTIONS")
    def print_default_o(self, id):
        return ""

    @appier.route("/nodes/<str:id>/print_hello", ("GET", "POST"), json=True)
    @appier.ensure(token="admin")
    def print_hello_default(self, id):
        self.set_field("data_b64", HELLO_WORLD_B64)
        self.set_field("name", "hello_world")
        return self.print_default(id)

    @appier.route("/nodes/<str:id>/printers/print", ("GET", "POST"), json=True)
    @appier.ensure(token="admin")
    def print_printer_f(self, id):
        printer = self.field("printer")
        return self.print_printer(id, printer)

    @appier.route("/nodes/<str:id>/printers/print", "OPTIONS")
    def print_printer_of(self, id):
        printer = self.field("printer")
        return self.print_printer_o(id, printer)

    @appier.route("/nodes/<str:id>/printers/print_hello", ("GET", "POST"), json=True)
    @appier.ensure(token="admin")
    def print_hello_printer_f(self, id):
        printer = self.field("printer")
        return self.print_hello_printer(id, printer)

    @appier.route(
        "/nodes/<str:id>/printers/<str:printer>/print", ("GET", "POST"), json=True
    )
    @appier.ensure(token="admin")
    def print_printer(self, id, printer):
        data = self.field("data", None)
        data_b64 = self.field("data_b64", None)
        name = self.field("name", None)
        type = self.field("type", None)
        format = self.field("format", None)
        options = self.field("options", None, cast=dict)

        appier.verify(
            data or data_b64,
            message="Either data or data_b64 fields must be provided",
            code=400,
        )
        appier.verify(
            not (data and data_b64),
            message="Only one of data or data_b64 fields must be provided",
            code=400,
        )

        job_id = str(uuid.uuid4())
        name = name or job_id
        if data:
            data_b64 = base64.b64encode(
                appier.legacy.bytes(data, encoding="utf-8")
            ).decode("utf-8")

        job_info = dict(
            id=job_id, name=name, node_id=id, printer=printer, data_length=len(data_b64)
        )
        if type:
            job_info["type"] = type
        if format:
            job_info["format"] = format
        if options:
            job_info["options"] = dict(
                (k, v) for k, v in options.items() if k in VALID_OPTIONS
            )
        self.owner.jobs_info[job_id] = job_info

        # creates a copy of the job info as starting
        # point for the job structure and then adds
        # the "heavy" data (base64 encoded) to it
        job = dict(job_info)
        job["data_b64"] = data_b64
        jobs = self.owner.jobs.get(id, [])
        jobs.append(job)
        self.owner.jobs[id] = jobs
        appier.notify("jobs:%s" % id)

        job_info.update(status="queued", queued_time=time.time())
        return job_info

    @appier.route("/nodes/<str:id>/printers/<str:printer>/print", "OPTIONS")
    def print_printer_o(self, id, printer):
        return ""

    @appier.route(
        "/nodes/<str:id>/printers/<str:printer>/print_hello", ("GET", "POST"), json=True
    )
    @appier.ensure(token="admin")
    def print_hello_printer(self, id, printer):
        self.set_field("data_b64", HELLO_WORLD_B64)
        self.set_field("name", "hello_world")
        return self.print_printer(id, printer)

    @appier.coroutine
    def wait_jobs(self, id):
        while True:
            jobs = self.owner.jobs.pop(id, [])
            if jobs:
                break
            for value in appier.wait("jobs:%s" % id):
                yield value
        yield json.dumps(jobs)
        for job in jobs:
            job_info = self.owner.jobs_info[job["id"]]
            job_info.update(status="printing", printing_time=time.time())



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import appier
import base64

import PIL.Image

from . import exceptions

from ..common.base import *
from ..manager.ast import *

try:
    import reportlab.lib.units
except ImportError:
    INCH = 72.0
    CM = INCH / 2.54
else:
    INCH = reportlab.lib.units.inch
    CM = reportlab.lib.units.cm

FONT_SCALE_FACTOR = 1
""" The font scale factor """

IMAGE_SCALE_FACTOR = 0.5
""" The image scale factor """

EXCLUSION_LIST = [
    "__class__",
    "__delattr__",
    "__dict__",
    "__doc__",
    "__getattribute__",
    "__hash__",
    "__init__",
    "__module__",
    "__new__",
    "__reduce__",
    "__reduce_ex__",
    "__repr__",
    "__setattr__",
    "__str__",
    "__weakref__",
    "__format__",
    "__sizeof__",
    "__subclasshook__",
    "accept",
    "accept_double",
    "accept_post_order",
    "add_child_node",
    "remove_child_node",
    "set_indent",
    "set_value",
    "indent",
    "value",
    "child_nodes",
]
""" The exclusion list """

FONT_SUFFIX_MAP = {"regular": "", "bold": "b", "italic": "i", "bold_italic": "z"}
""" The map associating the type of font
and the suffix to be appended to the name
to created the full font name """

DEFAULT_ENCODER = "utf-8"
""" The default encoder """

SCALE = CM
""" The scale value to be used in the conversion
of the centimeter value into the PDF point """

TWIP_SCALE = 0.001763889
""" The scale factor to convert a twip value into
a centimeter base value """

A4_PAPER = (21.0 * SCALE, 29.7 * SCALE)
""" The default size (dimensions) for an a4 paper
based structure, this includes the additional
margin values normally created by full size printers """

ROLL_PAPER = (7.2 * SCALE, 29.7 * SCALE)
""" The default size (dimensions) for a roll paper
based structure, this includes the additional
margin values normally created by receipt printers """

PAPER_SIZE = A4_PAPER
""" The default paper size to be used when no paper
size value is defined for the print operation """

FONT_PATHS = ("", "~/.fonts/", "/usr/share/fonts/truetype/")
""" The set of base paths to be used for searching
for fonts on the current system """


class Visitor(object):
    """
    The visitor class for the PDF printing infra-structure
    this visitor should be able to generate a PDF file
    using the reportlab library as basics.
    """

    node_method_map = {}
    """ The node method map """

    visit_childs = True
    """ The visit childs flag """

    visit_next = True
    """ The visit next flag """

    visit_index = 0
    """ The visit index, for multiple visits """

    options = {}
    """ The printing options """

    canvas = None
    """ The reference to the canvas object to be used
    for manipulating the various PDF elements """

    size = None
    """ The size of the document to be printed, this
    is the width and height measured in PDF points """

    width = None
    """ The width of the document to be printed, this
    is the width measured in PDF points """

    height = None
    """ The height of the document to be printed, this
    is the height measured in PDF points """

    single = False
    """ If the document to be parsed is considered to be
    single page only or multi page, this is relevant for
    the fixed/absolute positioning using blocks  """

    current_position = None
    """ The current position in the document measured
    as PDF points """

    fonts = {}
    """ The map containing the various loaded fonts
    to avoid multiple loading of fonts (redundancy is
    removed to avoid errors) """

    context_map = {}
    """ The context information map """

    def __init__(self):
        self.node_method_map = {}
        self.visit_childs = True
        self.visit_next = True
        self.visit_index = 0
        self.options = {}
        self.canvas = None
        self.width = 0
        self.height = 0
        self.single = False
        self.current_position = None
        self.fonts = {}
        self.context_map = {}

        self.update_node_method_map()

    def update_node_method_map(self):
        # retrieves the class of the current instance
        self_class = self.__class__

        # retrieves the names of the elements for the current class
        self_class_elements = dir(self_class)

        # iterates over all the name of the elements
        for self_class_element in self_class_elements:
            # retrieves the real element value
            self_class_real_element = getattr(self_class, self_class_element)

            # in case the current class real element does not contain
            # an AST node class reference must continue the loop
            if not hasattr(self_class_real_element, "ast_node_class"):
                continue

            # retrieves the AST node class from the current class real element
            # and sets it in the node method map
            ast_node_class = getattr(self_class_real_element, "ast_node_class")
            self.node_method_map[ast_node_class] = self_class_real_element

    def get_options(self):
        """
        Retrieves the printing options.

        :rtype: Dictionary
        :return: The printing options.
        """

        return self.options

    def set_options(self, options):
        """
        Sets the printing options.

        :type options: Dictionary.
        :param options: The printing options.
        """

        self.options = options

    @dispatch_visit()
    def visit(self, node):
        print("unrecognized element node of type " + node.__class__.__name__)

    def before_visit(self, node):
        self.visit_childs = True
        self.visit_next = True

    def after_visit(self, node):
        pass

    @visited(AstNode)
    def visit_ast_node(self, node):
        pass

    @visited(GenericElement)
    def visit_generic_element(self, node):
        pass

    @visited(PrintingDocument)
    def visit_printing_document(self, node):
        import reportlab.pdfgen.canvas

        # in case it's the first visit
        if self.visit_index == 0:
            # adds the node as the context information
            self.add_context(node)

            # retrieves both the file (buffer) to be used for the output
            # of the PDF file contents and the expected size for the PDF
            # document, in case no size is provided a default one is used
            file = self.options["file"]
            size = self.options.get("size", PAPER_SIZE)

            # tries to retrieve the document width and height values and
            # converts them into the proper integer values, these values
            # are mostly available for single page operations
            printing_document_width = hasattr(node, "width") and int(node.width) or 0
            printing_document_height = hasattr(node, "height") and int(node.height) or 0

            # in case both the printing document with and height values are
            # available the size tuple for the current document is defined
            # using their values, note that these values are defined as tenth
            # of the millimeter values (100 units equals 1 centimeter)
            if not printing_document_width == 0 and not printing_document_height == 0:
                size = (
                    printing_document_width / 100 * SCALE,
                    printing_document_height / 100 * SCALE,
                )

            # unpacks the size tuple into the width and height
            # components and sets the tuple containing both values
            # as the size to be used for the document
            self.width, self.height = size
            self.size = (self.width, self.height)

            # creates the canvas object to be used as the primary
            # entry point for operation on the PDF
            self.canvas = reportlab.pdfgen.canvas.Canvas(file, pagesize=self.size)

            # sets the initial position so that the "virtual" cursor
            # position is situated at the top left corner of the page
            self.current_position = (0, self.height)

        # in case it's the second visit
        elif self.visit_index == 1:
            # saves the final canvas structure flushing the data to
            # the associated file object, this is considered the final
            # operation for the creation of the PDF file
            self.canvas.save()

            # removes the context information
            self.remove_context(node)

    @visited(Block)
    def visit_block(self, node):
        if self.visit_index == 0:
            # adds the node as the context information, this way
            # the complete set of symbols for the block are exposed
            # to the underlying nodes (block opening)
            self.add_context(node)
            self.push_context("biggest_height", 0)

        # in case it's the second visit
        elif self.visit_index == 1:
            # removes the context information
            self.remove_context(node)

    @visited(Paragraph)
    def visit_paragraph(self, node):
        if self.visit_index == 0:
            self.add_context(node)
        elif self.visit_index == 1:
            self.remove_context(node)

    @visited(Line)
    def visit_line(self, node):
        if self.visit_index == 0:
            self.add_context(node)
            self.push_context("biggest_height", 0)

            # retrieves the margin top value defined
            # for the current context
            margin_top = int(self.get_context("margin_top", "0"))

            # retrieves the current position in x and y
            # and then updates the current position
            current_position_x, current_position_y = self.current_position
            self.current_position = (
                current_position_x,
                current_position_y - margin_top * FONT_SCALE_FACTOR,
            )

        elif self.visit_index == 1:
            biggest_height = self.get_context("biggest_height")
            self.pop_context("biggest_height")

            # retrieves the margin bottom value defined
            # for the current context
            margin_bottom = int(self.get_context("margin_bottom", "0"))

            # retrieves the current position in x and y
            # and then updates the current position
            current_position_x, current_position_y = self.current_position
            self.current_position = (
                0,
                current_position_y - biggest_height - margin_bottom * FONT_SCALE_FACTOR,
            )

            # removes the context information
            self.remove_context(node)

    @visited(Text)
    def visit_text(self, node):
        if self.visit_index == 0:
            # adds the node as the context information
            self.add_context(node)

            # retrieves the text and encodes it using
            # the default encoder and ignoring possible errors
            text_encoded = node.text.encode(DEFAULT_ENCODER, "ignore")

            # retrieves the complete set of attributes for the current
            # context to be used for the processing of the node
            font_name = str(self.get_context("font"))
            font_size = int(self.get_context("font_size"))
            text_align = self.get_context("text_align")
            font_style = self.get_context("font_style", "regular")
            margin_left = int(self.get_context("margin_left", "0"))
            margin_right = int(self.get_context("margin_right", "0"))
            position_x = int(self.get_context("x", "0"))
            position_y = int(self.get_context("y", "0"))
            block_width = int(self.get_context("width", "0"))
            block_height = int(self.get_context("height", "0"))

            # converts the various block related values from their original
            # twip based value into the PDF point value to be used in print
            position_x = int(position_x * TWIP_SCALE * SCALE)
            position_y = int(position_y * TWIP_SCALE * SCALE)
            block_width = int(block_width * TWIP_SCALE * SCALE)
            block_height = int(block_height * TWIP_SCALE * SCALE)

            # verifies if the current context is of type block, so that
            # the proper absolute positions are going to be used instead
            is_block = not block_width == 0 and not block_height == 0

            # sets the initial clip (box) value that will be applied in
            # case the current execution mode is not block based
            clip_left = 0
            clip_top = 0
            clip_right = self.width
            _clip_bottom = self.height

            # verifies if the current mode is block and if that's the case
            # re-calculates the new clip (box) value taking that into account
            if is_block:
                self.single = True
                clip_left = position_x
                clip_top = position_y * -1
                clip_right = position_x + block_width
                _clip_bottom = (position_y + block_height) * -1

            # calculates the resized font size so that it's normalized
            # according to the printing language specification, the value
            # is rounded to one decimal place so that no major visual
            # rounding problems occur
            font_size_r = round(font_size / 1.2, 1)

            # retrieves the proper suffix for the requested font style
            # and uses it to create the complete font name ensuring that
            # it's currently loaded in the PDF context
            suffix = FONT_SUFFIX_MAP.get(font_style, "")
            font_name_c = font_name + suffix
            self.ensure_font(font_name_c)

            # sets the complete computed font in the current canvas context
            # note that the leading value is overridden to avoid font sizing
            # problems in accordance with the printing language specification,
            # this is the "first" font setting and ensures that the measuring
            # of the font size is the correct one (required by algorithm)
            self.canvas.setFont(font_name_c, font_size_r, leading=1.0)

            # retrieves the current position in x and y unpacking the values
            # from the current position tuple (as expected)
            _current_position_x, current_position_y = self.current_position

            # calculates the text height from the font scale factor
            # and measures the text width using the underlying rendering
            # infra-structure (avoids possible problems)
            text_height = font_size * FONT_SCALE_FACTOR
            text_height_r = font_size_r * FONT_SCALE_FACTOR
            text_width = self.canvas.stringWidth(text_encoded)

            # initializes the text x coordinate with the margin defined
            # for the current node (difference of margins)
            text_x = (margin_left - margin_right) * FONT_SCALE_FACTOR

            # calculates the appropriate text position according to the
            # "requested" horizontal text alignment
            if text_align == "left":
                text_x += clip_left
            elif text_align == "right":
                text_x += clip_right - text_width
            elif text_align == "center":
                text_x += (
                    clip_left + int((clip_right - clip_left) / 2) - int(text_width / 2)
                )

            # sets the text y as the current position context y
            # default position for the text is the current position
            text_y = clip_top + current_position_y - text_height_r
            text_y = self.ensure_y(text_y, offset=text_height)

            # updates the fill color to a white color and then uses
            # this color to draw a white (background) rectangle around
            # the area that is going to be filled with the text
            self.canvas.setFillColorRGB(1.0, 1.0, 1.0)
            self.canvas.rect(
                text_x,
                text_y - (font_size - font_size_r),
                text_width,
                text_height,
                stroke=0,
                fill=1,
            )

            # "resets" the fill color of the current canvas context to
            # the black color so that it's possible to draw the text with
            # the expected color values as defined in specification
            self.canvas.setFillColorRGB(0, 0, 0)

            # sets the complete computed font in the current canvas context
            # note that the leading value is overridden to avoid font sizing
            # problems in accordance with the printing language specification,
            # the font must be set after the ensure vertical operation so that
            # in case a new page is created the new font is set correctly in it
            self.canvas.setFont(font_name_c, font_size_r, leading=1.0)

            # draws the text string at the calculated position the text
            # is encoded in the expected encoding so that no encoding
            # problems occur (as expected)
            self.canvas.drawString(text_x, text_y, text_encoded)

            # in case the current text height is bigger than the current
            # context biggest height, updates the information
            biggest_height = self.get_context("biggest_height")
            if biggest_height < text_height:
                self.put_context("biggest_height", text_height)

        # in case it's the second visit
        elif self.visit_index == 1:
            # removes the context information
            self.remove_context(node)

    @visited(Image)
    def visit_image(self, node):
        import reportlab.lib.utils

        if self.visit_index == 0:
            # adds the node as the context information
            self.add_context(node)

            # sets the default values for both the image path
            # and source, both values are unset by default
            image_path = None
            image_source = None

            # retrieves the path or source value to be used
            # in the retrieval (only one value is set)
            if self.has_context("path"):
                image_path = self.get_context("path")
            elif self.has_context("source"):
                image_source = self.get_context("source")

            # retrieves the complete set of attributes for the current
            # context to be used for the processing of the node
            text_align = self.get_context("text_align")
            position_x = int(self.get_context("x", "0"))
            position_y = int(self.get_context("y", "0"))
            block_width = int(self.get_context("width", "0"))
            block_height = int(self.get_context("height", "0"))

            # converts the various block related values from their original
            # twip based value into the PDF point value to be used in print
            position_x = int(position_x * TWIP_SCALE * SCALE)
            position_y = int(position_y * TWIP_SCALE * SCALE)
            block_width = int(block_width * TWIP_SCALE * SCALE)
            block_height = int(block_height * TWIP_SCALE * SCALE)

            # verifies if the current context is of type block, so that
            # the proper absolute positions are going to be used instead
            is_block = not block_width == 0 and not block_height == 0

            # sets the initial clip (box) value that will be applied in
            # case the current execution mode is not block based
            clip_left = 0
            clip_top = 0
            clip_right = self.width
            _clip_bottom = self.height

            # verifies if the current mode is block and if that's the case
            # re-calculates the new clip (box) value taking that into account
            if is_block:
                self.single = True
                clip_left = position_x
                clip_top = position_y * -1
                clip_right = position_x + block_width
                _clip_bottom = (position_y + block_height) * -1

            # in case the image path is defined must load the
            # image data from the file system
            if image_path:
                # opens the bitmap image directly from the current
                # file system, no dynamically loaded image
                bitmap_image = PIL.Image.open(image_path)

            # in case the image source is defined must load the
            # base 64 image data from the attribute
            elif image_source:
                # decodes the image source from the default base 64
                # encoding to be used for the loading
                image_source_decoded = base64.b64decode(image_source)

                # creates the image buffer then writes the decoded
                # image into it and opens the file object with the
                # created buffer (image loading into structure)
                image_source_buffer = appier.legacy.BytesIO()
                image_source_buffer.write(image_source_decoded)
                image_source_buffer.seek(0)
                bitmap_image = PIL.Image.open(image_source_buffer)

            # retrieves the bitmap image width and height
            bitmap_image_width, bitmap_image_height = bitmap_image.size

            # creates a new image without transparency settings, so that
            # no extra color is used ands copies the bitmap image into it
            other_image = PIL.Image.new(
                "RGB", (bitmap_image_width, bitmap_image_height), color="white"
            )
            other_image.paste(bitmap_image, bitmap_image)

            # retrieves the current position in x and y unpacking the values
            # from the current position tuple
            _current_position_x, current_position_y = self.current_position

            # calculates the appropriate bitmap position according to the
            # "requested" horizontal text alignment
            if text_align == "left":
                real_bitmap_x = clip_left
            elif text_align == "right":
                real_bitmap_x = clip_right - bitmap_image_width * IMAGE_SCALE_FACTOR
            elif text_align == "center":
                real_bitmap_x = (
                    clip_left
                    + int((clip_right - clip_left) / 2)
                    - int(bitmap_image_width * IMAGE_SCALE_FACTOR / 2)
                )

            # calculates the real bitmap vertical position from the current
            # vertical position minus the height of the image and ensures the
            # position, recalculating a new y position in case the page overflows
            real_bitmap_y = (
                clip_top
                + current_position_y
                - (bitmap_image_height * IMAGE_SCALE_FACTOR)
            )
            real_bitmap_y = self.ensure_y(
                real_bitmap_y, offset=bitmap_image_height * IMAGE_SCALE_FACTOR
            )

            # loads the image image using the proper image reader structure
            # and uses the structure to "draw" the image into the canvas at
            # the current position
            image_reader = reportlab.lib.utils.ImageReader(other_image)
            self.canvas.drawImage(
                image_reader,
                real_bitmap_x,
                real_bitmap_y,
                bitmap_image_width * IMAGE_SCALE_FACTOR,
                bitmap_image_height * IMAGE_SCALE_FACTOR,
            )

            # in case the current image height is bigger than the current
            # context biggest height, updates the information
            biggest_height = self.get_context("biggest_height")
            if biggest_height < bitmap_image_height * IMAGE_SCALE_FACTOR:
                self.put_context(
                    "biggest_height", bitmap_image_height * IMAGE_SCALE_FACTOR
                )

        elif self.visit_index == 1:
            self.remove_context(node)

    def ensure_y(self, y_position, offset):
        """
        Ensures that the provided vertical position is valid for
        the current page (no overflow) in case it's not a new page
        is created and the returned position is the new one.

        The calculation of the new vertical position takes into
        account the provided offset value.

        :type y_position: float
        :param y_position: The vertical position as a PDF point value
        to be validated against the current page metrics.
        :type offset: float
        :param offset: The vertical offset to be applied to the new's
        page vertical value in case a new page is created.
        :rtype: float
        :return: The resulting vertical position taking into account
        creation of new pages.
        """

        # in case the current document in parsing has been marked
        # as single page oriented (absolute positioning), there's
        # no need to verify for new page creation
        if self.single:
            return y_position

        # verifies if the current vertical position "overflows"
        # the page value (lower than zero) in case it does not
        # returns immediately with the provided position (no
        # overflow has occurred)
        b_position = y_position - offset
        if b_position >= 0.0:
            return y_position

        # calculates the delta position (inside the new page)
        # according to the current y position in case it's positive
        # it should be ignored otherwise uses it as the offset
        d_position = 0.0 if y_position >= 0.0 else y_position

        # updates the current position with the initial top left
        # corner position of the new page and update the vertical
        # position coordinate with the offset value and the bottom
        # position (used as an offset)
        self.current_position = (0, self.height + d_position)
        y_position = self.height + d_position - offset

        # shows a new page in the current canvas (creating a new
        # page) and then returns the new vertical position to the
        # caller method
        self.canvas.showPage()
        return y_position

    def ensure_font(self, font_name, file_path=None):
        """
        Ensures that the font is present in the current
        canvas object, loading it into the PDF context
        in case it's required.

        The provided file path may be an absolute path
        or a relative (file name) to the system's default
        font directory.

        :type font_name: String
        :param font_name: The name of the font to be ensured
        to be loaded in the current context.
        :type file_path: String
        :param file_path: The path to the true type font file
        that should be loaded for the font (may be a relative
        or absolute path)
        """

        import reportlab.pdfbase.ttfonts
        import reportlab.pdfbase.pdfmetrics

        # in case the font is already present in the fonts
        # map it's considered to be loaded and so the control
        # must be returned immediately
        if font_name in self.fonts:
            return

        # converts the font name into a lower cased version and
        # then uses it to create the default font path in case
        # none is provided (uses the default true type extension)
        font_name_l = font_name.lower()
        file_path = file_path or font_name_l + ".ttf"

        # sets the error flag as true by default the loading of
        # the font is not possible in case one of the steps
        # completes the error flag is unset
        error = True

        # iterates over all the font paths trying to find a path
        # that can correctly load the requested font
        for font_path in FONT_PATHS:
            try:
                # creates the complete font path with the current
                # base path in iteration and the font name in case
                # the font path is empty the default system wide
                # search will be used (varies from system to system)
                file_path_f = font_path + file_path
                file_path_f = os.path.expanduser(file_path_f)

                # creates the font structure for the font name and path
                # and the registers it in the the current report lab
                # metrics (to be used in further operations)
                font = reportlab.pdfbase.ttfonts.TTFont(font_name, file_path_f)
                reportlab.pdfbase.pdfmetrics.registerFont(font)
            except Exception:
                continue
            else:
                error = False
                break

        # in case the error flag is set raises the invalid font
        # exception, indicating that it was not possible to load
        # the associated font file
        if error:
            raise exceptions.InvalidFont(
                "not possible to load '%s' - '%s'" % (font_name, file_path)
            )

        # updates the fonts map so that the current font is associated
        # with the corresponding loaded file path, this marks the
        # font as loaded for the current context
        self.fonts[font_name] = file_path

    def get_current_position_context(self):
        """
        Retrieves the current position based on the current
        context information.

        :rtype: Tuple
        :return: The current position base on the current context information
        and applied with the font scale factor.
        """

        # retrieves the current position in x and y
        current_position_x, current_position_y = self.current_position

        # converts the current position to context
        current_position_context = (
            FONT_SCALE_FACTOR * current_position_x,
            -1 * FONT_SCALE_FACTOR * current_position_y,
        )

        # returns the current position context
        return current_position_context

    def get_context(self, context_name, default=None):
        if not self.has_context(context_name):
            if not default == None:
                return default
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        return self.peek_context(context_name)

    def add_context(self, node):
        valid_attributes = [
            (value, getattr(node, value))
            for value in dir(node)
            if not value in EXCLUSION_LIST
        ]

        for valid_attribute_name, valid_attribute_value in valid_attributes:
            self.push_context(valid_attribute_name, valid_attribute_value)

    def remove_context(self, node):
        valid_attribute_names = [
            value for value in dir(node) if not value in EXCLUSION_LIST
        ]

        for valid_attribute_name in valid_attribute_names:
            self.pop_context(valid_attribute_name)

    def push_context(self, context_name, context_value):
        if not context_name in self.context_map:
            self.context_map[context_name] = []

        self.context_map[context_name].append(context_value)

    def pop_context(self, context_name):
        if not context_name in self.context_map:
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        self.context_map[context_name].pop()

    def peek_context(self, context_name):
        if not context_name in self.context_map:
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        return self.context_map[context_name][-1]

    def put_context(self, context_name, context_value):
        """
        Puts the given context information in the context
        information map.

        :type context_name: String
        :param context_name: The name of the context information
        to be put in the context information map.
        :type context_value: Object
        :param context_value: The value of the context information to be put
        in the context information map.
        """

        if not context_name in self.context_map:
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        self.context_map[context_name][-1] = context_value

    def has_context(self, context_name):
        """
        Tests if the given context information name exists
        in the current context information map.

        :type context_name: String
        :param context_name: The context information name
        to be tested against the current context information map.
        :rtype: bool
        :return: If the context information name exists in the
        current context information map (and is valid).
        """

        # in case the context information name exists in the
        # context information map and is not invalid
        if context_name in self.context_map and self.context_map[context_name]:
            return True
        else:
            return False



#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier

import xml.dom.minidom

from . import ast

DEFAULT_CHARSET = "utf-8"
""" The default charset """


class Parser(object):
    """
    The abstract parser class
    """

    def __init__(self):
        """
        Constructor of the class
        """

        pass

    def parse(self):
        """
        Parses the defined file
        """

        pass

    def get_value(self):
        """
        Retrieves the result of the parse

        :rtype: Object
        :return: The result of the parse
        """

        pass

    def parse_element_attributes(self, node, element):
        attributes = node.attributes

        for index in range(attributes.length):
            attribute_node = attributes.item(index)
            attribute_node_name = attribute_node.name
            attribute_node_value = attribute_node.value
            setattr(element, attribute_node_name, attribute_node_value)


class PrintingLanguageParser(Parser):
    """
    The printing language parser class.
    """

    file = None
    """ The file path """

    string = None
    """ The string contents """

    printing_document = None
    """ The printing document """

    def __init__(self, file=None, string="none"):
        Parser.__init__(self)
        self.file = file
        self.string = string

    def parse(self):
        self.load_printing_language_file(self.file)

    def parse_string(self):
        self.load_printing_language_string(self.string)

    def get_value(self):
        return self.printing_document

    def get_build_automation(self):
        return self.printing_document

    def load_printing_language_file(self, file):
        # creates the XML document DOM object
        xml_document = xml.dom.minidom.parse(file)

        self.load_printing_language(xml_document)

    def load_printing_language_string(self, string):
        # creates the XML document DOM object, the provided
        # string is encoded to avoid possible parsing problems
        string = (
            type(string) == appier.legacy.UNICODE
            and string.encode(DEFAULT_CHARSET)
            or string
        )
        xml_document = xml.dom.minidom.parseString(string)

        self.load_printing_language(xml_document)

    def load_printing_language(self, xml_document):
        child_nodes = xml_document.childNodes

        for child_node in child_nodes:
            if valid_node(child_node):
                self.printing_document = self.parse_printing_document(child_node)

    def parse_printing_document(self, printing_document):
        printing_document_structure = ast.PrintingDocument()
        child_nodes = printing_document.childNodes

        # parses the element attributes
        self.parse_element_attributes(printing_document, printing_document_structure)

        for child_node in child_nodes:
            if valid_node(child_node):
                self.parse_printing_document_element(
                    child_node, printing_document_structure
                )

        return printing_document_structure

    def parse_printing_document_element(
        self, printing_document_element, printing_document
    ):
        node_name = printing_document_element.nodeName
        printing_document_child_nodes = printing_document.child_nodes

        if node_name == "block":
            printing_document_child_nodes.append(
                self.parse_block(printing_document_element)
            )
        elif node_name == "paragraph":
            printing_document_child_nodes.append(
                self.parse_paragraph(printing_document_element)
            )
        elif node_name == "line":
            printing_document_child_nodes.append(
                self.parse_line(printing_document_element)
            )
        elif node_name == "text":
            printing_document_child_nodes.append(
                self.parse_text(printing_document_element)
            )
        elif node_name == "image":
            printing_document_child_nodes.append(
                self.parse_image(printing_document_element)
            )

    def parse_block(self, block):
        block_structure = ast.Block()
        child_nodes = block.childNodes

        # parses the element attributes
        self.parse_element_attributes(block, block_structure)

        for child_node in child_nodes:
            if valid_node(child_node):
                self.parse_block_element(child_node, block_structure)

        return block_structure

    def parse_block_element(self, block_element, block):
        node_name = block_element.nodeName
        block_child_nodes = block.child_nodes

        if node_name == "text":
            block_child_nodes.append(self.parse_text(block_element))
        elif node_name == "image":
            block_child_nodes.append(self.parse_image(block_element))

    def parse_paragraph(self, paragraph):
        paragraph_structure = ast.Paragraph()
        child_nodes = paragraph.childNodes

        # parses the element attributes
        self.parse_element_attributes(paragraph, paragraph_structure)

        for child_node in child_nodes:
            if valid_node(child_node):
                self.parse_paragraph_element(child_node, paragraph_structure)

        return paragraph_structure

    def parse_paragraph_element(self, paragraph_element, paragraph):
        node_name = paragraph_element.nodeName
        paragraph_child_nodes = paragraph.child_nodes

        if node_name == "line":
            paragraph_child_nodes.append(self.parse_line(paragraph_element))
        elif node_name == "text":
            paragraph_child_nodes.append(self.parse_text(paragraph_element))
        elif node_name == "image":
            paragraph_child_nodes.append(self.parse_image(paragraph_element))

    def parse_line(self, line):
        line_structure = ast.Line()
        child_nodes = line.childNodes

        # parses the element attributes
        self.parse_element_attributes(line, line_structure)

        for child_node in child_nodes:
            if valid_node(child_node):
                self.parse_line_elements(child_node, line_structure)

        return line_structure

    def parse_line_elements(self, line_element, line):
        node_name = line_element.nodeName
        line_child_nodes = line.child_nodes

        if node_name == "text":
            line_child_nodes.append(self.parse_text(line_element))
        elif node_name == "image":
            line_child_nodes.append(self.parse_image(line_element))

    def parse_text(self, text):
        text_structure = ast.Text()

        # parses the element attributes
        self.parse_element_attributes(text, text_structure)

        if text.firstChild:
            text_structure.text = text.firstChild.data.strip()
        else:
            text_structure.text = str()

        return text_structure

    def parse_image(self, image):
        image_structure = ast.Image()

        # parses the element attributes
        self.parse_element_attributes(image, image_structure)

        return image_structure


def valid_node(node):
    """
    Gets if a node is valid or not for parsing.

    :type node: Node
    :param node: The Xml node to be validated.
    :rtype: bool
    :return: The valid or not valid value.
    """

    # in case the node is of type element
    if node.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
        # returns true (valid)
        return True
    # otherwise
    else:
        # returns false (invalid)
        return False



#!/usr/bin/python
# -*- coding: utf-8 -*-

from . import visitor

PRINTING_NAME = "pdf"
""" The printing name, for the current PDF
infra-structure (as defined in spec) """


class PrintingPDF(object):
    """
    The printing PDF class, that is responsible for
    the handling of the logic for the printing of PDF
    files based on the XML template language.
    """

    def get_name(self):
        return PRINTING_NAME

    def print_test(self, options={}):
        pass

    def print_test_image(self, image_path, options={}):
        pass

    def print_language(self, printing_document, options={}):
        # creates the PDF printing visitor then sets the
        # provided printing options in the visitor
        _visitor = visitor.Visitor()
        _visitor.set_options(options)

        # accepts the visitor in the printing document,
        # using double visiting mode
        printing_document.accept_double(_visitor)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from . import ast
from . import exceptions
from . import parser
from . import system
from . import visitor

from .ast import (
    AstNode,
    GenericElement,
    PrintingDocument,
    Block,
    Paragraph,
    Line,
    Text,
    Image,
)
from .exceptions import PrintingManagerException, PrintingPluginNotAvailable
from .parser import Parser, valid_node
from .system import PrintingManager
from .visitor import Visitor



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os

import appier

from . import parser
from . import exceptions

TEST_IMAGE_PATH = "resources/test_logo.png"
""" The test image relative path to the current
file path (considered the base path) """


class PrintingManager(object):
    """
    The printing manager class, that is responsible for the
    top level handling of the printing operations.
    """

    handlers_map = {}
    """ The map containing an association between the
    name of the printing handler and the proper instance
    that may be used for printing processing """

    def __init__(self):
        self.handlers_map = {}

    def load(self):
        import colony_print

        pdf_handler = colony_print.PrintingPDF()
        binie_handler = colony_print.PrintingBinie()

        self.load_handler(pdf_handler)
        self.load_handler(binie_handler)

    def unload(self):
        for _name, handler in appier.legacy.items(self.handlers_map):
            self.unload_handler(handler)

    def print_test(self, options={}):
        # retrieves the proper handler using the provided map of options
        # and then uses the same options to run a test print operation
        handler = self._get_handler(options)
        handler.print_test(options)

    def print_test_image(self, options={}):
        # retrieves the complete path for the current file and then
        # retrieves it's directory path, to be used in the calculus
        # of the image path to be used
        file_path = os.path.realpath(__file__)
        base_path = os.path.dirname(file_path)

        # creates the complete image path using the calculated base
        # path from the current file's path and then appends the relative
        # path to the image resource
        image_path = os.path.join(base_path, TEST_IMAGE_PATH)

        # retrieves the proper printing handler for the provided options
        # and then uses it to print the test image in the calculated path
        handler = self._get_handler(options)
        handler.print_test_image(image_path, options)

    def print_language(self, data, options={}):
        # creates a new printing language parser and sets the
        # proper data in it running then the parse string operation
        # that is going to be parsing the provided string
        _parser = parser.PrintingLanguageParser()
        _parser.string = data
        _parser.parse_string()

        # retrieves the (printing) document resulting from
        # the parsing of the provided value this value is
        # going to be sent to the handler for "printing"
        document = _parser.get_value()

        # retrieves the proper handler according to the provided
        # options and uses it in the printing operation
        handler = self._get_handler(options)
        handler.print_language(document, options)

    def load_handler(self, handler):
        # retrieves the printing name from the handler and
        # uses it to register the handler in the proper map
        printing_name = handler.get_name()
        self.handlers_map[printing_name] = handler

    def unload_handler(self, handler):
        # gathers the name of the provided handler and then
        # removes any reference to it in the map of handlers
        printing_name = handler.get_name()
        del self.handlers_map[printing_name]

    def _get_handler(self, options):
        # retrieves the printing name (engine) from the printing options
        # this value is going to be used to select the proper handler
        printing_name = options.get("name", None)
        if not printing_name:
            raise exceptions.PrintingPluginNotAvailable("missing name")

        # tries to retrieve the proper handler for the requested name that
        # exists in the handlers map in case it's not available raises an
        # exception indicating the problem
        handler = self.handlers_map.get(printing_name, None)
        if not handler:
            raise exceptions.PrintingPluginNotAvailable("no handler for requested name")

        # returns the (printing) handler that has been requested according
        # to the provided map of options
        return handler



#!/usr/bin/python
# -*- coding: utf-8 -*-

from . import binie
from . import common
from . import manager
from . import pdf

from .binie import *
from .common import *
from .manager import *
from .pdf import *



#!/usr/bin/python
# -*- coding: utf-8 -*-


class PrintingManagerException(Exception):
    """
    The printing manager exception class.
    """

    message = None
    """ The exception's message """


class PrintingPluginNotAvailable(PrintingManagerException):
    """
    The printing plugin not available name class.
    """

    def __init__(self, message):
        """
        Constructor of the class.

        :type message: String
        :param message: The message to be printed.
        """

        PrintingManagerException.__init__(self)
        self.message = message

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        return "Printing plugin not available - %s" % self.message



#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier


class JobController(appier.Controller):
    @appier.route("/jobs", "GET", json=True)
    @appier.ensure(token="admin")
    def list(self):
        return dict(self.owner.jobs_info)

    @appier.route("/jobs/<str:id>", "GET", json=True)
    @appier.ensure(token="admin")
    def show(self, id):
        return self.owner.jobs_info[id]



#!/usr/bin/python
# -*- coding: utf-8 -*-

from . import base
from . import document
from . import job
from . import node
from . import printer

from .base import BaseController
from .document import DocumentController
from .job import JobController
from .node import NodeController
from .printer import PrinterController



#!/usr/bin/python
# -*- coding: utf-8 -*-


def visited(ast_node_class):
    def decorator(function, *args, **kwargs):
        function.ast_node_class = ast_node_class
        return function

    return decorator


def dispatch_visit(map_name="node_method_map"):
    def create_interceptor(function):
        def decorator_interceptor(*args, **kwargs):
            # unpacks the first two "unnamed" arguments as the self
            # instance reference and the node element to be visited
            self = args[0]
            node = args[1]

            # verifies if the current instance contains the node method
            # map if that's the case retrieves otherwise falls back to an
            # empty dictionary (for code compatibility)
            has_map = hasattr(self, map_name)
            node_method_map = self.node_method_map if has_map else dict()

            # retrieves the class for the node argument and then
            # gathers the complete mro class definition to be able
            # to iterate over the class hierarchy
            node_class = node.__class__
            mro = node_class.mro()

            # iterates over the complete class hierarchy for the provided
            # node (from bottom to up) so that the best match for the
            # visit operation is found and properly called
            for mro_item in mro:
                # in case the current mro item class level is nor found
                # skips the current iteration (cannot visit at this level)
                if not mro_item in node_method_map:
                    continue

                # the current class level is valid and so the proper method
                # is retrieved from the map and then called with the provided
                # arguments, note that a before visit and an after visit calls
                # are done so that proper "notification" exists
                visit_method = node_method_map[mro_item]
                self.before_visit(*args[1:], **kwargs)
                visit_method(*args, **kwargs)
                self.after_visit(*args[1:], **kwargs)

                # returns the control flow to the caller method, note that no
                # value is returned as this is a simple visit operation
                return

            # in case no visit has been made a normal fallback operation is
            # performed by calling the "original" function/method
            return function(*args, **kwargs)

        return decorator_interceptor

    def decorator(function, *args, **kwargs):
        interceptor = create_interceptor(function)
        return interceptor

    return decorator



#!/usr/bin/python
# -*- coding: utf-8 -*-


class AstNode(object):
    """
    The AST node class, that represents a generic node
    to be percolated as part of the AST (abstract syntax tree).
    """

    value = None
    """ The value """

    indent = False
    """ The indentation level """

    child_nodes = []
    """ The list of child nodes """

    def __init__(self):
        self.child_nodes = []

    def __repr__(self):
        return "<ast_node indent:%s child_nodes:%s>" % (
            self.indent,
            len(self.child_nodes),
        )

    def accept(self, visitor):
        visitor.visit(self)

        if visitor.visit_childs:
            for child_node in self.child_nodes:
                child_node.accept(visitor)

    def accept_post_order(self, visitor):
        if visitor.visit_childs:
            for child_node in self.child_nodes:
                child_node.accept_post_order(visitor)

        visitor.visit(self)

    def accept_double(self, visitor):
        visitor.visit_index = 0
        visitor.visit(self)

        if visitor.visit_childs:
            for child_node in self.child_nodes:
                child_node.accept_double(visitor)

        visitor.visit_index = 1
        visitor.visit(self)

    def set_value(self, value):
        self.value = value

    def set_indent(self, indent):
        self.indent = indent

    def add_child_node(self, child_node):
        self.child_nodes.append(child_node)

    def remove_child_node(self, child_node):
        self.child_nodes.remove(child_node)


class GenericElement(AstNode):
    element_name = "none"

    def __init__(self, element_name="none"):
        AstNode.__init__(self)
        self.element_name = element_name


class PrintingDocument(AstNode):
    def __init__(self):
        AstNode.__init__(self)


class Block(AstNode):
    def __init__(self):
        AstNode.__init__(self)


class Paragraph(AstNode):
    def __init__(self):
        AstNode.__init__(self)


class Line(AstNode):
    def __init__(self):
        AstNode.__init__(self)


class Text(AstNode):
    def __init__(self):
        AstNode.__init__(self)


class Image(AstNode):
    def __init__(self):
        AstNode.__init__(self)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import struct
import base64
import appier

import PIL.Image

from . import exceptions

from ..common.base import *
from ..manager.ast import *

FONT_SCALE_FACTOR = 20
""" The font scale factor """

IMAGE_SCALE_FACTOR = 10
""" The image scale factor """

EXCLUSION_LIST = [
    "__class__",
    "__delattr__",
    "__dict__",
    "__doc__",
    "__getattribute__",
    "__hash__",
    "__init__",
    "__module__",
    "__new__",
    "__reduce__",
    "__reduce_ex__",
    "__repr__",
    "__setattr__",
    "__str__",
    "__weakref__",
    "__format__",
    "__sizeof__",
    "__subclasshook__",
    "accept",
    "accept_double",
    "accept_post_order",
    "add_child_node",
    "remove_child_node",
    "set_indent",
    "set_value",
    "indent",
    "value",
    "child_nodes",
]
""" The exclusion list """

DEFAULT_ENCODER = "utf-8"
""" The default encoder """


class Visitor(object):
    """
    The visitor class.
    """

    node_method_map = {}
    """ The node method map """

    visit_childs = True
    """ The visit childs flag """

    visit_next = True
    """ The visit next flag """

    visit_index = 0
    """ The visit index, for multiple visits """

    options = {}
    """ The printing options """

    elements_list = []
    """ The list containing the various elements """

    current_position = None
    """ The current position """

    context_map = {}
    """ The context information map """

    def __init__(self):
        self.node_method_map = {}
        self.visit_childs = True
        self.visit_next = True
        self.visit_index = 0
        self.options = {}
        self.elements_list = []
        self.current_position = None
        self.context_map = {}

        self.update_node_method_map()

    def update_node_method_map(self):
        # retrieves the class of the current instance
        self_class = self.__class__

        # retrieves the names of the elements for the current class
        self_class_elements = dir(self_class)

        # iterates over all the name of the elements
        for self_class_element in self_class_elements:
            # retrieves the real element value
            self_class_real_element = getattr(self_class, self_class_element)

            # in case the current class real element does not contain
            # an AST node class reference must continue the loop
            if not hasattr(self_class_real_element, "ast_node_class"):
                continue

            # retrieves the AST node class from the current class real element
            # and sets it in the node method map
            ast_node_class = getattr(self_class_real_element, "ast_node_class")
            self.node_method_map[ast_node_class] = self_class_real_element

    def get_options(self):
        """
        Retrieves the printing options.

        :rtype: Dictionary
        :return: The printing options.
        """

        return self.options

    def set_options(self, options):
        """
        Sets the printing options.

        :type options: Dictionary.
        :param options: The printing options.
        """

        self.options = options

    @dispatch_visit()
    def visit(self, node):
        print("unrecognized element node of type " + node.__class__.__name__)

    def before_visit(self, node):
        self.visit_childs = True
        self.visit_next = True

    def after_visit(self, node):
        pass

    @visited(AstNode)
    def visit_ast_node(self, node):
        pass

    @visited(GenericElement)
    def visit_generic_element(self, node):
        pass

    @visited(PrintingDocument)
    def visit_printing_document(self, node):
        # in case it's the first visit
        if self.visit_index == 0:
            # adds the node as the context information
            self.add_context(node)

            # retrieves the printing document name
            printing_document_name = node.name

            # resets the list of elements in the document
            self.elements_list = []

            # sets the initial position
            self.current_position = (0, 0)
        # in case it's the second visit
        elif self.visit_index == 1:
            # retrieves the printing document name and dimensions
            # to be able to update the structure
            printing_document_name = node.name
            printing_document_width = hasattr(node, "width") and int(node.width) or 0
            printing_document_height = hasattr(node, "height") and int(node.height) or 0

            # ensures that a proper byte based representation exists
            # for the current printing document name (as expected)
            printing_document_name = str(printing_document_name)
            printing_document_name = appier.legacy.bytes(printing_document_name)

            # packs the header value as a binary string
            header = struct.pack(
                "<256sIII",
                printing_document_name,
                printing_document_width,
                printing_document_height,
                len(self.elements_list),
            )
            self.options["file"].write(header)

            # iterates over all the elements list to create their header
            # and then add the element data
            for element in self.elements_list:
                # unpacks the element into the type and its
                # data (contents)
                element_type, element_data = element
                element_length = len(element_data)

                # creates the element header using the type and the length
                # of the element and then writes it and the data into the
                # the current file
                element_header = struct.pack("<II", element_type, element_length)
                self.options["file"].write(element_header)
                self.options["file"].write(element_data)

            # removes the context information
            self.remove_context(node)

    @visited(Block)
    def visit_block(self, node):
        if self.visit_index == 0:
            # adds the node as the context information, this way
            # the complete set of symbols for the block are exposed
            # to the underlying nodes (block opening)
            self.add_context(node)
            self.push_context("biggest_height", 0)

        # in case it's the second visit
        elif self.visit_index == 1:
            # removes the context information
            self.remove_context(node)

    @visited(Paragraph)
    def visit_paragraph(self, node):
        if self.visit_index == 0:
            self.add_context(node)
        elif self.visit_index == 1:
            self.remove_context(node)

    @visited(Line)
    def visit_line(self, node):
        if self.visit_index == 0:
            self.add_context(node)
            self.push_context("biggest_height", 0)

            # retrieves the margin top value defined
            # for the current context
            margin_top = int(self.get_context("margin_top", "0"))

            # retrieves the current position in x and y
            # and then updates the current position
            current_position_x, current_position_y = self.current_position
            self.current_position = (
                current_position_x,
                current_position_y - margin_top * FONT_SCALE_FACTOR,
            )

        elif self.visit_index == 1:
            biggest_height = self.get_context("biggest_height")
            self.pop_context("biggest_height")

            # retrieves the margin bottom value defined
            # for the current context
            margin_bottom = int(self.get_context("margin_bottom", "0"))

            # retrieves the current position in x and y
            # and then updates the current position
            current_position_x, current_position_y = self.current_position
            self.current_position = (
                0,
                current_position_y - biggest_height - margin_bottom * FONT_SCALE_FACTOR,
            )

            # removes the context information
            self.remove_context(node)

    @visited(Text)
    def visit_text(self, node):
        if self.visit_index == 0:
            # adds the node as the context information
            self.add_context(node)

            # retrieves the text and encodes it using
            # the default encoder and ignoring possible errors
            text_encoded = node.text.encode(DEFAULT_ENCODER, "ignore")

            # retrieves the complete set of attributes for the current
            # context to be used for the processing of the node
            font_name = str(self.get_context("font"))
            font_size = int(self.get_context("font_size"))
            text_align = self.get_context("text_align")
            font_style = self.get_context("font_style", "regular")
            margin_left = int(self.get_context("margin_left", "0"))
            margin_right = int(self.get_context("margin_right", "0"))
            position_x = int(self.get_context("x", "0"))
            position_y = int(self.get_context("y", "0"))
            block_width = int(self.get_context("width", "0"))
            block_height = int(self.get_context("height", "0"))

            # ensures that the string text values are represented as
            # byte oriented objects as expected by specification
            font_name = appier.legacy.bytes(font_name)
            font_name = appier.legacy.bytes(font_name)

            # sets the default values for the text weight and for
            # the italic enumeration
            text_weight_int = 0
            text_italic_int = 0

            if font_style == "bold" or font_style == "bold_italic":
                text_weight_int = 1

            if font_style == "italic" or font_style == "bold_italic":
                text_italic_int = 1

            # retrieves the current position in x and y
            _current_position_x, _current_position_y = self.current_position

            # converts the provided text align value into the
            # appropriate integer value representing it
            if text_align == "left":
                text_align_int = 1
            elif text_align == "right":
                text_align_int = 2
            elif text_align == "center":
                text_align_int = 3

            # calculates the text height from the font scale factor
            text_height = font_size * FONT_SCALE_FACTOR

            # packs the element text element structure containing all the meta
            # information that makes part of it then adds the "just" created
            # element to the elements list
            element = struct.pack(
                "<ii256sIIIIIIIIIII",
                0,
                _current_position_y,
                font_name,
                font_size,
                text_align_int,
                text_weight_int,
                text_italic_int,
                margin_left,
                margin_right,
                position_x,
                position_y,
                block_width,
                block_height,
                len(text_encoded) + 1,
            )
            element += text_encoded
            element += b"\0"
            self.elements_list.append((1, element))

            # in case the current text height is bigger than the current
            # context biggest height, updates the information
            biggest_height = self.get_context("biggest_height")
            if biggest_height < text_height:
                self.put_context("biggest_height", text_height)

        # in case it's the second visit
        elif self.visit_index == 1:
            # removes the context information
            self.remove_context(node)

    @visited(Image)
    def visit_image(self, node):
        if self.visit_index == 0:
            # adds the node as the context information
            self.add_context(node)

            # sets the default values for both the image path
            # and source, both values are unset by default
            image_path = None
            image_source = None

            # retrieves the path or source value to be used
            # in the retrieval (only one value is set)
            if self.has_context("path"):
                image_path = self.get_context("path")
            elif self.has_context("source"):
                image_source = self.get_context("source")

            # retrieves the complete set of attributes for the current
            # context to be used for the processing of the node
            text_align = self.get_context("text_align")
            position_x = int(self.get_context("x", "0"))
            position_y = int(self.get_context("y", "0"))
            block_width = int(self.get_context("width", "0"))
            block_height = int(self.get_context("height", "0"))

            # in case the image path is defined must load the
            # image data from the file system
            if image_path:
                # opens the bitmap image directly from the current
                # file system, no dynamically loaded image
                bitmap_image = PIL.Image.open(image_path)

            # in case the image source is defined must load the
            # base 64 image data from the attribute
            elif image_source:
                # decodes the image source from the default base 64
                # encoding to be used for the loading
                image_source_decoded = base64.b64decode(image_source)

                # creates the image buffer then writes the decoded
                # image into it and opens the file object with the
                # created buffer (image loading into structure)
                image_source_buffer = appier.legacy.BytesIO()
                image_source_buffer.write(image_source_decoded)
                image_source_buffer.seek(0)
                bitmap_image = PIL.Image.open(image_source_buffer)

            # retrieves the bitmap image width and height
            bitmap_image_width, bitmap_image_height = bitmap_image.size

            # creates a new image without transparency settings, so that
            # no extra color is used ands copies the bitmap image into it
            other_image = PIL.Image.new(
                "RGB", (bitmap_image_width, bitmap_image_height), color="white"
            )
            other_image.paste(bitmap_image, bitmap_image)

            # retrieves the current position in x and y
            _current_position_x, current_position_y = self.current_position

            # converts the provided text align value into the
            # appropriate integer value representing it
            if text_align == "left":
                text_align_int = 1
            elif text_align == "right":
                text_align_int = 2
            elif text_align == "center":
                text_align_int = 3

            # sets the real bitmap image height as the bitmap
            # image height (value copy)
            real_bitmap_image_height = bitmap_image_height

            # creates a new string buffer for the image
            string_buffer = appier.legacy.BytesIO()

            # saves the new image into the string buffer and then
            # retrieve the buffer data
            other_image.save(string_buffer, "bmp")
            buffer = string_buffer.getvalue()

            # packs the element image element structure containing all the meta
            # information that makes part of it then adds the "just" created
            # element to the elements list
            element = struct.pack(
                "<iiIIIIII",
                0,
                current_position_y,
                text_align_int,
                position_x,
                position_y,
                block_width,
                block_height,
                len(buffer),
            )
            element += buffer
            self.elements_list.append((2, element))

            biggest_height = self.get_context("biggest_height")
            if biggest_height < real_bitmap_image_height * IMAGE_SCALE_FACTOR:
                self.put_context(
                    "biggest_height", real_bitmap_image_height * IMAGE_SCALE_FACTOR
                )

        elif self.visit_index == 1:
            self.remove_context(node)

    def get_current_position_context(self):
        """
        Retrieves the current position based on the current
        context information.

        :rtype: Tuple
        :return: The current position base on the current context information
        and applied with the font scale factor.
        """

        # retrieves the current position in x and y
        current_position_x, current_position_y = self.current_position

        # converts the current position to context
        current_position_context = (
            FONT_SCALE_FACTOR * current_position_x,
            -1 * FONT_SCALE_FACTOR * current_position_y,
        )

        # returns the current position context
        return current_position_context

    def get_context(self, context_name, default=None):
        if not self.has_context(context_name):
            if not default == None:
                return default
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        return self.peek_context(context_name)

    def add_context(self, node):
        valid_attributes = [
            (value, getattr(node, value))
            for value in dir(node)
            if value not in EXCLUSION_LIST
        ]

        for valid_attribute_name, valid_attribute_value in valid_attributes:
            self.push_context(valid_attribute_name, valid_attribute_value)

    def remove_context(self, node):
        valid_attribute_names = [
            value for value in dir(node) if value not in EXCLUSION_LIST
        ]

        for valid_attribute_name in valid_attribute_names:
            self.pop_context(valid_attribute_name)

    def push_context(self, context_name, context_value):
        if not context_name in self.context_map:
            self.context_map[context_name] = []

        self.context_map[context_name].append(context_value)

    def pop_context(self, context_name):
        if not context_name in self.context_map:
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        self.context_map[context_name].pop()

    def peek_context(self, context_name):
        if not context_name in self.context_map:
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        return self.context_map[context_name][-1]

    def put_context(self, context_name, context_value):
        """
        Puts the given context information in the context
        information map.

        :type context_name: String
        :param context_name: The name of the context information
        to be put in the context information map.
        :type context_value: Object
        :param context_value: The value of the context information to be put
        in the context information map.
        """

        if not context_name in self.context_map:
            raise exceptions.InvalidContextInformationName(
                "the context information name: " + context_name + " is invalid"
            )

        self.context_map[context_name][-1] = context_value

    def has_context(self, context_name):
        """
        Tests if the given context information name exists
        in the current context information map.

        :type context_name: String
        :param context_name: The context information name
        to be tested against the current context information map.
        :rtype: bool
        :return: If the context information name exists in the
        current context information map (and is valid).
        """

        # in case the context information name exists in the
        # context information map and is not invalid
        if context_name in self.context_map and self.context_map[context_name]:
            return True
        else:
            return False



#!/usr/bin/python
# -*- coding: utf-8 -*-

import base64

import appier

MIME = dict(binie="text/x-binie", pdf="application/pdf")
""" Map defining the association between the print
format naming and the associated base mime type value
(note that this value may be complemented with base64) """

EXAMPLE = '<?xml version="1.0" encoding="UTF-8"?>\
    <printing_document name="hello_world" font="Calibri" font_size="9">\
        <paragraph text_align="center">\
            <line><text>Hello World</text></line>\
        </paragraph>\
    </printing_document>'
""" Example XML string that should display an hello world
message using the XML printing language (XMPL) """


class DocumentController(appier.Controller):
    def __init__(self, owner, *args, **kwargs):
        appier.Controller.__init__(self, owner, *args, **kwargs)
        self.manager = None

    @appier.route("/documents/example.<format>", "GET")
    def example(self, format):
        return self.send_print(EXAMPLE, format)

    @appier.route("/documents.<format>", "POST")
    def convert(self, format):
        # retrieves the current request reference and then
        # uses it to retrieve its "raw" data, which should
        # contain the XML string (XMPL) for the generation of the
        # of binie result and then sends the value for print
        request = self.get_request()
        data = request.get_data()
        return self.send_print(data, format=format)

    def send_print(self, data, format="binie"):
        """
        Converts ("prints") the provided data (XMPL) into the target
        format, either Binie or PDF, and then sends the converted
        data back to the client.

        This method is expected to be running inside an Appier request
        handling context, as it sets content type and other headers.

        :type data: String
        :param data: The data to be converted into the target format.
        :type format: String
        :param format: The target format to be used for the conversion.
        :rtype: String
        :return: The converted data in the target format.
        :see: https://github.com/hivesolutions/colony-print/blob/master/doc/xmpl.md
        :see: https://github.com/hivesolutions/colony-print/blob/master/doc/binie.md
        """

        # retrieves the various optional fields for printing
        # and then parses them creating the composite values
        # (should include the size tuple)
        b64 = self.field("base64", False, cast=bool)
        width = self.field("width", 0.0, cast=float)
        height = self.field("height", 0.0, cast=float)
        has_size = width > 0.0 and width > 0.0

        mime = self.get_mime(format, b64=base64)
        manager = self.get_manager()

        data = data
        file = appier.legacy.BytesIO()
        options = dict(name=format, file=file)
        if has_size:
            options["size"] = (width, height)

        manager.print_language(data, options)
        value = file.getvalue()
        value = base64.b64encode(value) if b64 else value

        self.content_type(mime)
        return value

    def get_mime(self, format, b64=False):
        mime = MIME.get(format, "application/octet-stream")
        mime = mime + "-base64" if b64 else mime
        return mime

    def get_manager(self):
        import colony_print

        if self.manager:
            return self.manager
        self.manager = colony_print.PrintingManager()
        self.manager.load()
        return self.manager



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import setuptools

setuptools.setup(
    name="colony-print",
    version="0.4.3",
    author="Hive Solutions Lda.",
    author_email="development@hive.pt",
    description="Colony Print Infra-structure",
    license="Apache License, Version 2.0",
    keywords="colony print native",
    url="http://colony-print.hive.pt",
    zip_safe=False,
    packages=[
        "colony_print",
        "colony_print.controllers",
        "colony_print.printing",
        "colony_print.printing.binie",
        "colony_print.printing.common",
        "colony_print.printing.manager",
        "colony_print.printing.pdf",
    ],
    test_suite="colony_print.test",
    package_dir={"": os.path.normpath("src")},
    package_data={
        "colony_print": [
            "static/example/*",
            "static/example/js/*",
            "static/example/xml/*",
        ]
    },
    install_requires=["appier", "appier-extras", "jinja2", "pillow", "reportlab"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    long_description=open(os.path.join(os.path.dirname(__file__), "README.md"), "rb")
    .read()
    .decode("utf-8"),
    long_description_content_type="text/markdown",
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import time
import uuid
import json
import base64
import shutil
import logging
import tempfile

import appier

NAME = "colony-print"
""" The name of the program currently running """

VERSION = "0.4.3"
""" The version of the colony print node currently running,
this value should be updated whenever a new version is released """

BASE_URL = "https://print.bemisc.com/"
""" The default base URL to be used for the communication with the
Colony Print server """

SLEEP_TIME = 3.0
""" The default time to sleep between each iteration, this value
is used to avoid overloading the server with requests """

NODE_MODES = set(["normal", "email"])
""" The set of running modes that are considered to be valid for
the node, this is going to be used to validate the mode """

EMAIL_TEMPLATE = appier.legacy.u(
    """
Hey there!

Great news — your document **%s** has just gone through a virtual transformation and is now rocking the PDF stage! 🎸📄 Ready to take a look? Check out the attachment—it's dressed to impress.

We hope you find everything in perfect harmony. Should you need a replay, just hit 'print' again!

Keep on printing,

Your's dear 'Colony Print'

P.S. No trees were harmed in the making of this PDF. 🌳✌️
"""
)


class ColonyPrintNode(object):
    def __init__(self, sleep_time=SLEEP_TIME):
        self.sleep_time = sleep_time
        self.node_mode = None
        self.node_printer = None
        self.node_email_receivers = None

    def loop(self):
        logging.basicConfig(
            format="%(asctime)s [%(levelname)s] %(message)s", level=logging.DEBUG
        )

        base_url = appier.conf("BASE_URL", BASE_URL)
        secret_key = appier.conf("SECRET_KEY", None)
        node_id = appier.conf("NODE_ID", "node")
        node_name = appier.conf("NODE_NAME", "node")
        node_location = appier.conf("NODE_LOCATION", "undefined")
        self.node_mode = appier.conf("NODE_MODE", "normal")
        self.node_printer = appier.conf("NODE_PRINTER", "default")
        self.node_email_receivers = appier.conf("NODE_EMAIL_RECEIVER", [], cast=list)
        self.node_email_receivers = appier.conf(
            "NODE_EMAIL_RECEIVERS", self.node_email_receivers, cast=list
        )

        logging.info("Booting %s %s (%s)" % (NAME, VERSION, appier.PLATFORM))
        logging.info("Running node '%s' in '%s' mode" % (node_id, self.node_mode))

        headers = dict()
        if secret_key:
            headers["X-Secret-Key"] = secret_key

        while True:
            try:
                logging.info("Submitting node information")
                appier.post(
                    base_url + "nodes/%s" % node_id,
                    data_j=dict(
                        name=node_name,
                        location=node_location,
                        mode=self.node_mode,
                        printer=self.node_printer,
                        engines=self.engines,
                        engine_info=self.engine_info,
                        platform=appier.PLATFORM,
                        os=os.name,
                        version=VERSION,
                    ),
                    headers=headers,
                )
                logging.info("Retrieving jobs for node '%s'" % node_id)
                jobs = appier.get(
                    base_url + "nodes/%s/jobs" % node_id, headers=headers, timeout=600
                )
                logging.info("Retrieved %d jobs for node '%s'" % (len(jobs), node_id))
                results = dict()
                for job in jobs:
                    try:
                        result = self.print_job(job)
                    except Exception as exception:
                        logging.exception(
                            "Exception while printing job '%s': %s"
                            % (job["id"], str(exception))
                        )
                        result = dict(result="error", error=str(exception))
                    results[job["id"]] = result
                for job_id, result in results.items():
                    logging.info("Posting job result for '%s'" % job_id)
                    appier.post(
                        base_url + "nodes/%s/jobs/%s/result" % (node_id, job_id),
                        data_j=result,
                        headers=headers,
                    )
            except Exception as exception:
                logging.exception("Exception while looping '%s'" % str(exception))
                logging.info("Sleeping for %.2f seconds" % self.sleep_time)
                time.sleep(self.sleep_time)

    def print_job(self, job):
        if not self.node_mode in NODE_MODES:
            raise appier.OperationalError("Mode '%s' not valid" % self.node_mode)
        return getattr(self, "print_job_" + self.node_mode)(job)

    def print_job_normal(self, job):
        return self._handle_job(job)

    def print_job_email(self, job):
        import mailme

        data_b64 = job["data_b64"]
        name = job.get("name", "undefined")
        printer = job.get("printer", None)
        format = job.get("format", None)
        options = job.get("options", dict())
        printer_s = printer if printer else self.node_printer
        short_name = name[-12:]

        self._ensure_format(format)

        temp_dir = tempfile.mkdtemp()
        try:
            logging.debug(
                "Created temporary directory '%s' for document generation" % temp_dir
            )

            output_path = os.path.join(temp_dir, "%s.pdf" % str(uuid.uuid4()))
            options["output_path"] = output_path
            logging.info(
                "Generating document job '%s' with '%s' printer" % (name, printer_s)
            )

            # sends the print job for handling using npcolony, this will make
            # sure that the job is printed in the current system
            self._handle_npcolony(
                data_b64, format=format, printer=printer_s, options=options
            )

            # does some busy waiting for the output file to be created
            # note that the process of handling the PDF printing is
            # asynchronous and may take some time to be completed
            for _ in range(10):
                if os.path.exists(output_path):
                    break
                time.sleep(0.5)

            file = open(output_path, "rb")
            try:
                data = file.read()
            finally:
                file.close()

            # computes the complete list of email receivers using the
            # base instance value and the ones provided via options,
            # makes sure that the email receivers are unique
            email_receivers = options.get("email_receivers", [])
            email_receiver = options.get("email_address", None)
            email_receiver = options.get("email_receiver", email_receiver)
            email_override = options.get("email_override", True)
            if email_receiver:
                email_receivers.append(email_receiver)
            email_receivers = (
                email_receivers
                if email_override
                else list(self.node_email_receivers) + email_receivers
            )

            logging.info(
                "Sending email to %s for job '%s' with '%s' printer"
                % (",".join(email_receivers), name, printer_s)
            )

            # creates the mailme API instance and sends the email with
            # the generated PDF file as attachment to the email receivers
            api = mailme.API()
            api.send(
                mailme.MessagePayload(
                    receivers=email_receivers,
                    title="Your PDF Masterpiece Awaits!",
                    subject="Print Job %s is Ready!" % short_name,
                    contents=appier.legacy.bytes(
                        EMAIL_TEMPLATE % name, encoding="utf-8", force=True
                    ),
                    attachments=[
                        mailme.AttachmentPayload(
                            name="%s.pdf" % name,
                            data=base64.b64encode(data).decode(),
                            mime="application/pdf",
                        )
                    ],
                )
            )
        finally:
            shutil.rmtree(temp_dir)

        return dict(
            result="success",
            mode="email",
            handler="npcolony",
            email_sent=True,
            receivers=email_receivers,
        )

    @property
    def npcolony(self):
        import npcolony

        return npcolony

    @property
    def engines(self):
        engines = []
        if self._has_npcolony():
            engines.append("npcolony")
        if self._has_gravo():
            engines.append("gravo")
        if self._has_text():
            engines.append("text")
        return engines

    @property
    def engine_info(self):
        engine_info = dict()
        if self._has_npcolony() and hasattr(self, "_info_npcolony"):
            engine_info["colony"] = self._info_npcolony()
        if self._has_gravo() and hasattr(self, "_info_gravo"):
            engine_info["gravo"] = self._info_gravo()
        if self._has_text() and hasattr(self, "_info_text"):
            engine_info["text"] = self._info_text()
        return engine_info

    def _handle_job(self, job):
        # unpacks the complete set of job information to
        # be able to print the job in the current system
        data_b64 = job["data_b64"]
        name = job.get("name", "undefined")
        printer = job.get("printer", None)
        type = job.get("type", None)
        format = job.get("format", None)
        options = job.get("options", dict())
        printer_s = printer if printer else self.node_printer

        logging.info("Printing job '%s' with '%s' printer" % (name, printer_s))
        if format:
            logging.info("Using format '%s' for job '%s'" % (format, name))

        if type in (None, "npcolony"):
            result = self._handle_npcolony(
                data_b64, format=format, printer=printer_s, options=options
            )
            return dict(
                result="success", handler="npcolony", printer=printer_s, data=result
            )
        elif type in ("gravo",):
            result = self._handle_gravo(data_b64)
            return dict(result="success", handler="gravo", data=result)
        elif type in ("text",):
            result = self._handle_text(data_b64)
            return dict(result="success", handler="text", data=result)

    def _handle_npcolony(self, data_b64, format=None, printer=None, options=dict()):
        if not self._has_npcolony():
            raise appier.OperationalError("npcolony engine is not available")

        self._ensure_format(format)

        if printer:
            self.npcolony.print_printer_base64(printer, data_b64, options=options)
        else:
            self.npcolony.print_base64(data_b64)

        return dict()

    def _handle_gravo(self, data_b64):
        if not self._has_gravo():
            raise appier.OperationalError("gravo engine is not available")

        import gravo_pilot

        data = base64.b64decode(data_b64)
        data_j = json.loads(data)

        text = data_j["text"]
        font = data_j.get("font", "HELVETICA 1L")
        font_size = data_j.get("font_size", None)
        dry_run = data_j.get("dry_run", False)

        start = time.time()
        screenshots = gravo_pilot.GravostyleAPI().write_text(
            text, font=font, font_size=font_size, dry_run=dry_run
        )
        duration = time.time() - start

        files = []

        for screenshot in screenshots:
            name, image = screenshot
            buffer = appier.legacy.BytesIO()
            image.save(buffer, format="PNG")
            data = buffer.getvalue()
            _data_b64 = base64.b64encode(data)
            files.append(appier.File(dict(name=name, data=_data_b64)).json_v())

        return dict(duration=duration, files=files)

    def _handle_text(self, data_b64):
        if not self._has_text():
            raise appier.OperationalError("text engine is not available")

        return dict(
            files=[appier.File(dict(name="document.txt", data=data_b64)).json_v()]
        )

    def _has_npcolony(self):
        try:
            __import__("npcolony")
        except Exception:
            return False
        return True

    def _has_gravo(self):
        try:
            __import__("gravo_pilot")
        except Exception:
            return False
        return True

    def _has_text(self):
        return True

    def _info_npcolony(self):
        return dict(
            format=self.npcolony.get_format(), devices=self.npcolony.get_devices()
        )

    def _ensure_format(self, format):
        # tries to make sure that the format is compatible with the current
        # system, this is required to avoid problems with the printing of the
        # data in printers of the current system
        if (
            format
            and hasattr(self.npcolony, "get_format")
            and not format == self.npcolony.get_format()
        ):
            raise appier.OperationalError(
                "Format '%s' not compatible with system" % format
            )


if __name__ == "__main__":
    node = ColonyPrintNode()
    node.loop()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier
import appier_extras


class ColonyPrintApp(appier.APIApp):
    def __init__(self, *args, **kwargs):
        appier.APIApp.__init__(
            self, name="colony-print", parts=(appier_extras.AdminPart,), *args, **kwargs
        )
        self.nodes = dict()
        self.jobs = dict()
        self.jobs_info = appier.LimitedSizeDict(
            max_size=appier.conf("JOB_SIZE", 1024, cast=int)
        )

    def _version(self):
        return "0.4.3"

    def _description(self):
        return "Colony Print"

    def _observations(self):
        return "Printing in the cloud"


if __name__ == "__main__":
    app = ColonyPrintApp()
    app.serve()
else:
    __path__ = []
